import torch
import torch.nn as nn

from .util import get_device


class TSCLoss(nn.Module):
    """
    Temporal Segmented Classification Loss
    """

    def __init__(self, falloff: str = 'kernel'):
        super(TSCLoss, self).__init__()

        self.falloff: str = falloff

    def forward(self, inputs, targets):
        loss: torch.Tensor = torch.tensor(0)

        if self.falloff == 'sinusoidal':
            loss = TSCLoss._generate_sinus_decay(targets, inputs.size()[1]) - inputs

        if self.falloff == 'kernel':
            loss = TSCLoss._generate_smoothing_kernel_decay(targets, inputs.size()[1]) - inputs

        return torch.mean(torch.pow(loss, 2))

    @staticmethod
    def _generate_smoothing_kernel_decay(targets: torch.Tensor, target_length: int, sigma: float = 4.0):
        def get_kernel(peak_index: int):
            kernel = torch.exp(
                -((torch.arange(0, target_length, dtype=torch.float32, device=get_device()) - peak_index) ** 2)
                / (2 * sigma ** 2)
            )
            kernel = kernel / kernel.sum()

            return kernel

        return torch.stack([get_kernel(tgt) for tgt in targets])

    @staticmethod
    def _generate_sinus_decay(
            targets: torch.Tensor,
            target_length: int,
            fall_off_minimum: float = 0.0,
            fall_off_power: int = 4,
    ) -> torch.Tensor:
        step_size = torch.pi / target_length

        return torch.pow(
            torch.maximum(
                torch.sin(
                    torch.stack([
                        torch.arange(
                            step_size / 2 + step_size * (target_length / 2 - tgt.item()),
                            torch.pi + step_size / 2 + step_size * (target_length / 2 - tgt.item()),
                            step_size,
                            device=get_device()
                        ) for tgt in targets
                    ])
                ),
                torch.tensor(fall_off_minimum)
            ),
            fall_off_power
        )



from typing import Union, Dict
from dataclasses import dataclass

import pandas as pd
from sklearn.metrics import f1_score, accuracy_score, classification_report


@dataclass
class Metric:
    decoding_fn: callable = lambda x: x
    golds: pd.Series = pd.Series([], dtype=object)
    preds: pd.Series = pd.Series([], dtype=object)

    def add_observations(self, golds: pd.Series, preds: pd.Series):
        self.golds = pd.concat([self.golds, golds], ignore_index=True)
        self.preds = pd.concat([self.preds, preds], ignore_index=True)

    def f_score(self, reduce: str = "weighted") -> float:
        return f1_score(self.golds, self.preds, average=reduce, zero_division=0.0)

    def accuracy(self) -> float:
        return accuracy_score(self.golds, self.preds)

    def cross_tabulation(self) -> pd.DataFrame:
        return pd.crosstab(
            self.golds.apply(self.decoding_fn),
            self.preds.apply(self.decoding_fn),
            rownames=['gold'],
            colnames=['pred']
        )

    def classification_report(self) -> pd.DataFrame:
        return pd.DataFrame(
            classification_report(
                self.golds.apply(self.decoding_fn),
                self.preds.apply(self.decoding_fn),
                zero_division=0.0,
                output_dict=True
            )
        ).T.drop('accuracy')

    def export(self, path: str) -> None:
        self.classification_report().to_csv(f'{path}/metric.classification_report.csv')
        self.cross_tabulation().to_csv(f'{path}/metric.cross_tabulation.csv')

    @property
    def data(self) -> Dict[str, pd.Series]:
        return {'golds': self.golds, 'preds': self.preds}

    def __repr__(self) -> Union[str, dict]:
        return self.classification_report().__repr__()



import logging as logger
import os
from typing import List, Optional, Tuple

import torch
from transformers import AutoTokenizer, AutoModel, logging

from .util import get_device, unpad, timing, model_memory_usage

os.environ["TOKENIZERS_PARALLELISM"] = "false"


class Encoder:

    @timing
    def __init__(
            self,
            model: str = "deepset/gbert-base",
            max_length: int = 512,
            layers: Optional[List[int]] = None,
    ):
        logging.set_verbosity_error()

        if layers is None:
            layers = [-1]

        self.tokenizer = AutoTokenizer.from_pretrained(model)
        self.model = AutoModel.from_pretrained(model, output_hidden_states=True).to(get_device())

        self.max_length = max_length
        self.layers = layers

        logger.info(self)

    @torch.no_grad()
    def __call__(
            self,
            batch: List[str],
            return_unpad: bool = True
    ) -> dict:
        token, encoding = self.tokenize(batch)
        embeds: torch.Tensor = self.forward(
            torch.tensor(encoding['input_ids'], device=get_device()).long(),
            torch.tensor(encoding['attention_mask'], device=get_device()).short()
        )

        return {
            'embeds': unpad(embeds, torch.tensor(encoding['attention_mask']).sum(1)) if return_unpad else embeds,
            'token': unpad(token, torch.tensor(encoding['attention_mask']).sum(1)) if return_unpad else token,
            **encoding
        }

    def tokenize(self, batch: List[str], padding: bool = True) -> Tuple[List[List[str]], dict]:
        encoding = self.tokenizer(
            batch,
            padding=padding,
            truncation=True,
            max_length=self.max_length
        )

        return (
            [self.ids_to_tokens(ids) for ids in encoding['input_ids']],
            encoding
        )

    @torch.no_grad()
    def forward(self, ids: torch.Tensor, masks: torch.Tensor) -> torch.Tensor:
        return torch.stack([
            self.model.forward(ids, masks).hidden_states[i]
            for i in self.layers
        ]).sum(0).squeeze()

    def ids_to_tokens(self, ids: torch.Tensor) -> List[str]:
        return self.tokenizer.convert_ids_to_tokens(ids)

    def ids_to_sent(self, ids: torch.Tensor) -> str:
        return self.tokenizer.decode(ids, skip_special_tokens=True)

    @property
    def dim(self) -> int:
        return self.model.config.to_dict()['hidden_size']

    def __len__(self) -> int:
        return self.model.config.to_dict()['vocab_size']

    def __repr__(self):
        return (
            f'> Encoder Name: {self.model.config.__dict__["_name_or_path"]}\n'
            f'  Memory Usage: {model_memory_usage(self.model)}'
        )



import argparse
import logging
from functools import wraps
from time import time
from typing import Union, Callable, Dict

import tomli
import torch


def setup_args_parser(description: str):
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        'config',
        metavar='C',
        type=str,
        help='path to config.toml file'
    )
    parser.add_argument(
        '--debug',
        default=False,
        action=argparse.BooleanOptionalAction,
        help='enable debug level logging'
    )
    return parser, parser.parse_args()


def setup_logging(debug: bool = False) -> None:
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.INFO,
        format="%(message)s",
        handlers=[logging.StreamHandler()]
    )


def load_config(path: str) -> Dict[str, any]:
    return tomli.load(open(path, 'rb'))


def get_device() -> str:
    return f'cuda' if torch.cuda.is_available() else 'cpu'


def unpad(
        padded: Union[list, torch.Tensor],
        length: Union[list, torch.Tensor]
) -> Union[list, torch.Tensor]:
    return [v[:n] for v, n in zip(padded, length)]


def timing(func: Callable):
    @wraps(func)
    def wrap(*args, **kwargs) -> any:
        start = time()
        result = func(*args, **kwargs)
        logging.info(f'> f({func.__name__}) took: {time() - start:2.4f} sec')

        return result

    return wrap


def model_memory_usage(model: torch.nn.Module) -> str:
    usage_in_byte: int = sum([
        sum([param.nelement() * param.element_size() for param in model.parameters()]),
        sum([buf.nelement() * buf.element_size() for buf in model.buffers()])
    ])

    return f'{usage_in_byte / (1024.0 * 1024.0):2.4f} MB'



import logging
from datetime import datetime
from typing import Tuple, Dict, Callable

import pandas as pd
import torch
from torch.optim import AdamW
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import get_scheduler

from .dataset import Dataset
from .metric import Metric


class Trainer:

    def __init__(
            self,
            dataset: Dict[str, Dataset],
            model: torch.nn.Module,
            collation_fn: Callable,
            config: dict
    ):
        self.dataset = dataset
        self.model = model
        self.collation_fn = collation_fn
        self.config: dict = config

        # progress tracking
        self.progress: dict = {
            'epoch': [],
            'loss_train': [],
            'loss_eval': [],
            'f1_train': [],
            'f1_eval': [],
            'duration': []
        }

        #  load optimizer and scheduler
        self.optimizer = AdamW(self.model.parameters(), lr=self.config['learning_rate'])
        self.scheduler = get_scheduler(
            "linear", optimizer=self.optimizer, num_warmup_steps=0,
            num_training_steps=self.config['num_epochs'] * len(self.dataset['train']),
        )

    def __call__(self) -> None:
        best_eval_metric: Dict[str, pd.Series] = {}

        try:
            for epoch in range(self.config['num_epochs']):
                _, metric_eval = self._epoch()
                self._log()

                if self.progress['f1_eval'][-1] == max(self.progress['f1_eval']):
                    best_eval_metric = metric_eval
                    self.model.save_pretrained(self.config["export_path"])

        except KeyboardInterrupt:
            logging.warning('Warning: Training interrupted by user, skipping to evaluation if possible.')

        if best_eval_metric:
            self._evaluate(best_eval_metric)

    def _epoch(self) -> Tuple[Dict[str, pd.Series], Dict[str, pd.Series]]:
        time_begin: datetime = datetime.now()

        loss_train, f1_train, metric_train = self._step(self.dataset['train'], optimize=True)
        loss_eval, f1_eval, metric_eval = self._step(self.dataset['val'])

        self.progress['epoch'].append(len(self.progress['epoch']) + 1)
        self.progress['duration'].append(datetime.now() - time_begin)

        self.progress['loss_train'].append(loss_train)
        self.progress['loss_eval'].append(loss_eval)

        self.progress['f1_train'].append(f1_train)
        self.progress['f1_eval'].append(f1_eval)

        return metric_train, metric_eval

    def _step(self, dataset: Dataset, optimize: bool = False) -> Tuple[float, float, Dict[str, pd.Series]]:
        loss_value: float = 0.0
        metric = Metric(decoding_fn=dataset.decode_target_label)

        for batch in tqdm(
                DataLoader(
                    dataset,
                    shuffle=True,
                    drop_last=True,
                    batch_size=self.config['batch_size'],
                    collate_fn=self.collation_fn
                ),
                leave=False
        ):

            loss = self._forward(batch, metric)
            loss_value += loss.item()

            if optimize:
                self._optimize(loss)

            del loss

        return loss_value / (len(dataset) / self.config['batch_size']), metric.f_score(), metric.data

    def _forward(self, batch: dict, metric: Metric) -> torch.Tensor:
        predictions, loss = self.model(**batch)

        metric.add_observations(
            pd.Series(batch['labels'].cpu().numpy()),
            pd.Series(torch.argmax(predictions, dim=1).cpu().numpy())
        )

        return loss

    def _optimize(self, loss: torch.Tensor) -> None:
        loss.backward()
        self.optimizer.step()
        self.scheduler.step()
        self.optimizer.zero_grad()

    def _evaluate(self, metric: Dict[str, pd.Series]) -> None:
        logging.info(f'[--- EVALUATION on max(f1_eval) ---]')
        metric = Metric(decoding_fn=self.dataset['train'].decode_target_label, **metric)
        metric.export(self.config["export_path"])
        logging.info(metric)

        (
            pd.DataFrame
            .from_records(self.progress, index=["epoch"])
            .to_csv(f'{self.config["export_path"]}/metric.train.csv')
        )

    def _log(self) -> None:
        logging.info((
            f'[@{self.progress["epoch"][-1]:03}]: \t'
            f'loss_train={self.progress["loss_train"][-1]:2.4f} \t'
            f'loss_eval={self.progress["loss_eval"][-1]:2.4f} \t'
            f'f1_train={self.progress["f1_train"][-1]:2.4f} \t'
            f'f1_eval={self.progress["f1_eval"][-1]:2.4f} \t'
            f'duration={self.progress["duration"][-1]}'
        ))



import logging

import torch
import torch.nn as nn

from .tscloss import TSCLoss
from .util import get_device, model_memory_usage


class Classifier(nn.Module):

    def __init__(
            self,
            in_size: int,
            out_size: int,
            hid_size: int = 32,
            dropout: float = 0.2,
            loss_fn: str = 'CrossEntropyLoss'
    ):
        super().__init__()

        self.model = nn.Sequential(
            nn.Linear(in_size, hid_size),
            nn.Dropout(p=dropout),
            nn.LeakyReLU(),
            nn.Linear(hid_size, out_size)
        ).to(get_device())

        if loss_fn == 'TSCLoss':
            self.loss = TSCLoss()

        else:
            self.loss = torch.nn.CrossEntropyLoss()

        logging.info(self)

    def __call__(self, embeds: torch.Tensor, labels: torch.Tensor):
        pred = self.predict(embeds)
        return pred, self.loss(pred, labels.view(-1))

    def predict(self, embeds: torch.Tensor) -> torch.Tensor:
        return self.model(embeds)

    def save_pretrained(self, path: str) -> None:
        torch.save({"state_dict": self.state_dict()}, f'{path}/model')

    def __len__(self) -> int:
        return sum(p.numel() for p in self.parameters() if p.requires_grad)

    def __repr__(self) -> str:
        return (
            f'> Model: Multilayer perceptron\n'
            f'  Memory Usage: {model_memory_usage(self.model)}'
            f' {self.model.__repr__()[:-1].replace("Sequential(", "")}'
        )



from . import Pipeline
from .util import setup_args_parser, setup_logging


class CLI:

    def __init__(self):
        parser, args = setup_args_parser('ProSem - Span Classification')

        setup_logging(args.debug)
        self.pipeline = Pipeline(args.config)

    def __call__(self) -> None:
        self.pipeline()


if __name__ == '__main__':
    CLI()()



"""ToDo!"""

__version__ = "0.1.1"

from .encoder import Encoder
from .classifier import Classifier
from .dataset import Dataset
from .trainer import Trainer
from .metric import Metric
from .pooler import Pooler

import logging
from typing import List

import pandas as pd
import torch

from .util import setup_args_parser, setup_logging, load_config, get_device


class Pipeline:

    def __init__(self, config: dict):
        self.config: dict = config

        logging.info('[--- LOAD ENCODER ---]')
        self.encoder = Encoder(**self.config['encoder'])
        self.target_mapping = {
            class_name: n for n, class_name in
            enumerate(self.config["dataset"]["label_classes"])
        }

        logging.info('[--- LOAD TRAINER ---]')
        self.trainer = Trainer(
            dataset={
                split: Dataset(
                    split=split,
                    data=pd.read_parquet(
                        f'{self.config["dataset"]["path"]}/{split}.parquet'
                    ),
                    data_label=self.config["dataset"]["text_column"],
                    target_label=self.config["dataset"]["label_column"],
                    target_mapping=self.target_mapping
                )
                for split in ['train', 'val']
            },
            model=self.load_classifier(),
            collation_fn=self.collate,
            config=self.config['trainer']
        )

    def __call__(self) -> None:
        logging.info('[--- RUN TRAINER ---]')
        self.trainer()

    def load_classifier(self) -> Classifier:
        return Classifier(
            in_size=self.encoder.dim,
            out_size=len(self.config['dataset']['label_classes']),
            **self.config['classifier']
        )

    def collate(self, batch: List[pd.DataFrame]) -> dict:
        collated_data: pd.DataFrame = pd.concat(batch)

        pre_collated: dict = {
            'labels': torch.tensor(
                [
                    self.target_mapping[label] for label
                    in collated_data[self.config["dataset"]["label_column"]].tolist()
                ],
                dtype=torch.long,
                device=get_device()
            ),
            **self.encoder(collated_data[self.config["dataset"]["text_column"]].tolist())
        }

        if 'subword' in self.config['pooler']['form']:
            pre_collated['text'] = collated_data[self.config["dataset"]["text_column"]].tolist()
            pre_collated['span_idx'] = collated_data[self.config["pooler"]["span_column"]].tolist()

        return {
            'embeds': Pooler.batch_pool(
                pre_collated,
                form=self.config['pooler']['form'],
                encoder=self.encoder
            ),
            'labels': pre_collated['labels']
        }



import logging
from dataclasses import dataclass
from typing import Dict, Union

import pandas as pd
from torch.utils.data import Dataset as torchDataset


@dataclass
class Dataset(torchDataset):
    split: str
    data: pd.DataFrame
    data_label: str
    target_label: str
    target_mapping: Dict[str, int]

    def __post_init__(self):
        logging.info(self)

    def encode_target_label(self, label: Union[str, bool]) -> int:
        return self.target_mapping.get(str(label))

    def decode_target_label(self, label: int) -> str:
        return {v: k for k, v in self.target_mapping.items()}.get(label)

    def __getitem__(self, idx) -> pd.DataFrame:
        return self.data.iloc[[idx]]

    def __len__(self) -> int:
        return len(self.data)

    def __repr__(self) -> str:
        return (
            f'> Dataset: {self.split}\n'
            f'  Samples: {len(self)}'
        )



from typing import Literal, List

import numpy as np
import torch

from .encoder import Encoder
from .util import get_device


class Pooler:
    extractions: dict = {
        # sentence based
        'cls': lambda x: x[0],
        'sent_mean': lambda x: torch.mean(x[1:-1], dim=0),

        # word based, positional extraction
        'subword_first': lambda x: x[0],
        'subword_last': lambda x: x[-1],

        # word based, arithmetic extraction
        'subword_mean': lambda x: torch.mean(x, dim=0),
        'subword_min': lambda x: torch.min(x, dim=0)[0],
        'subword_max': lambda x: torch.max(x, dim=0)[0]
    }

    @staticmethod
    def batch_pool(
            encoded_batch: dict,
            form: Literal[
                'cls', 'sent_mean',
                'subword_first', 'subword_last',
                'subword_mean', 'subword_min', 'subword_max'
            ] = 'cls',
            encoder: Encoder = None
    ):

        if form in ['cls', 'sent_mean']:
            return torch.stack([
                Pooler.extractions[form](embed)
                for embed in encoded_batch['embeds']
            ])

        else:
            span_text: List[str] = Pooler._extract_span_text(encoded_batch['text'], encoded_batch['span_idx'])
            span_token, _ = encoder.tokenize(span_text, padding=False)

            matching_embeds = Pooler._slice_span_matching_embeds(
                span_embeds=Pooler._extract_token_matching_embeds(
                    sent_embeds=encoded_batch['embeds'],
                    sent_token=encoded_batch['token'],
                    span_token=span_token
                ),
                span_position=Pooler._calculate_span_position(
                    span_frequency=Pooler._count_span_frequency(span_text, encoded_batch['text']),
                    span_text=span_text,
                    span_idx=encoded_batch['span_idx'],
                    text=encoded_batch['text']
                ),
                span_token=span_token
            )

            return torch.stack([Pooler.extractions[form](embed) for embed in matching_embeds])

    @staticmethod
    def _extract_span_text(text: List[str], spans: List[np.ndarray]) -> List[str]:
        return [
            ' '.join(sent.split()[slice(*span.astype(int).tolist())])
            for sent, span in zip(text, spans)
        ]

    @staticmethod
    def _count_span_frequency(span: List[str], text: List[str]) -> List[int]:
        return [sent.count(span) for span, sent in zip(span, text)]

    @staticmethod
    def _calculate_span_position(
            span_frequency: List[int],
            span_text: List[str],
            span_idx: List[int],
            text: List[str]
    ) -> List[int]:

        return [
            Pooler._find_idx(tx, span, idx[0]) if freq > 1 else -1
            for freq, span, idx, tx in zip(span_frequency, span_text, span_idx, text)
        ]

    @staticmethod
    def _find_idx(tx: str, span: str, idx: str | int) -> int:
        try:
            return [i for i in range(len(tx.split())) if tx.split()[i] == span].index(int(idx))

        except ValueError:
            return -1

    @staticmethod
    def _extract_token_matching_embeds(
            sent_embeds: List[torch.Tensor],
            sent_token: List[List[str]],
            span_token: List[List[str]],
    ) -> List[torch.Tensor]:
        return [
            embeds[torch.tensor(
                [tok in span[1:-1] for tok in token],
                device=get_device()
            ), :]
            for embeds, token, span in zip(
                sent_embeds,
                sent_token,
                span_token,
            )
        ]

    @staticmethod
    def _slice_span_matching_embeds(
            span_embeds: List[torch.Tensor],
            span_position: List[int],
            span_token: List[List[str]]
    ) -> List[torch.Tensor]:
        return [
            embed[position * size: position * size + size] if position != -1 else embed
            for embed, position, size in zip(
                span_embeds,
                span_position,
                [len(tok) - 2 for tok in span_token]
            )
        ]
